import contextlib
import io
import json
import logging
import os
import shlex
import subprocess
import tarfile
import time

import docker
import paramiko
from six.moves import shlex_quote

LOG = logging.getLogger("pubtools.quay")


# Docker client is called differently based on the version used. Unify the calls.
# Only use old call if version 1.X.X is used
class APIClient(docker.APIClient if int(docker.__version__[0]) > 1 else docker.client.Client):
    """Unify the call of Docker client for old and new version."""

    pass


# Python 2.6 version of paramiko doesn't support the usage
# of SSHClient as a context manager. This wrapper adds the functionality
@contextlib.contextmanager
def open_ssh_client():
    """Use SSHClient as a context manager."""
    client = paramiko.client.SSHClient()
    try:
        yield client
    finally:
        client.close()


class Executor(object):
    """
    Base executor class.

    Implementation of command execution should be done in
    descendant classes. Common pre- and post-processing should be
    implemented in this class.
    """

    def __enter__(self):
        """Use the class as context manager. Returns instance upon invocation."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Cleanup when used as context manager. No-op by default."""
        pass

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """Run a bash command."""
        raise NotImplementedError  # pragma: no cover"

    def skopeo_login(self, username=None, password=None):
        """
        Attempt to login to Quay if no login credentials are present.

        Args:
            username (str):
                Username for login.
            password (str):
                Password for login.
        """
        cmd_check = "skopeo login --get-login quay.io"
        out, err = self._run_cmd(cmd_check, tolerate_err=True)
        if username and username in out:
            LOG.info("Already logged in to Quay.io")
            return

        if not username or not password:
            raise ValueError(
                "Skopeo login credentials are not present. Quay user and token must be provided."
            )
        LOG.info("Logging in to Quay with provided credentials")

        cmd_login = (
            "skopeo login --authfile $HOME/.docker/config.json -u {0} --password-stdin quay.io"
        ).format(shlex_quote(username))
        out, err = self._run_cmd(cmd_login, stdin=password)

        if "Login Succeeded" in out:
            LOG.info("Login successful")
        else:
            raise RuntimeError(
                "Login command didn't generate expected output. "
                "STDOUT: '{0}', STDERR: '{1}'".format(out, err)
            )

    def tag_images(self, source_ref, dest_refs, all_arch=False):
        """
        Copy image from source to destination(s) using skopeo.

        Args:
            source_ref (str):
                Reference of the source image.
            dest_refs ([str]):
                List of target references to copy the image to.
            all_arch (bool):
                Whether to copy all architectures (if multiarch image)
        """
        if all_arch:
            cmd = "skopeo copy --all docker://{0} docker://{1}"
        else:
            cmd = "skopeo copy docker://{0} docker://{1}"

        for dest_ref in dest_refs:
            LOG.info("Tagging source '{0}' to destination '{1}'".format(source_ref, dest_ref))
            self._run_cmd(cmd.format(shlex_quote(source_ref), shlex_quote(dest_ref)))
            LOG.info("Destination image {0} has been tagged.".format(dest_ref))

        LOG.info("Tagging complete.")

    def skopeo_inspect(self, image_ref, raw=False):
        """
        Run skopeo inspect and return the result.

        NOTE: inspect command will not be run with the --raw argument. This option only returns an
        image manifest, which can be gathered by QuayClient. 'raw' argument in this function
        denotes if the result should be parsed or returned raw.

        Args:
            image_ref (str):
                Image reference to inspect.
            raw (bool):
                Whether to parse the returned JSON, or return raw.
        Returns (dict|str):
            Result of skopeo inspect.
        """
        cmd = "skopeo inspect docker://{0}".format(image_ref)
        out, _ = self._run_cmd(cmd)

        if raw:
            return out
        else:
            return json.loads(out)


class LocalExecutor(Executor):
    """Run commands locally."""

    def __init__(self, params={}):
        """
        Initialize.

        Args:
            params (dict):
                Custom parameters to be applied when running the shell commands.
        """
        self.params = params
        self.params.setdefault("universal_newlines", True)
        self.params.setdefault("stderr", subprocess.PIPE)
        self.params.setdefault("stdout", subprocess.PIPE)
        self.params.setdefault("stdin", subprocess.PIPE)

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """
        Run a command locally.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (str):
                String to send to standard input for a command.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."

        p = subprocess.Popen(shlex.split(cmd), **self.params)
        out, err = p.communicate(input=stdin)

        if p.returncode != 0 and not tolerate_err:
            LOG.error("Command {0} failed with {1}".format(cmd, err))
            raise RuntimeError(err_msg)

        return out, err


class RemoteExecutor(Executor):
    """Run commands remotely via SSH."""

    def __init__(
        self,
        hostname,
        username=None,
        key_filename=None,
        password=None,
        port=None,
        accept_unknown_host=True,
    ):
        """
        Initialize.

        Args:
            hostname (str):
                Host to connect to.
            username (str):
                Username to authenticate as. Defaults to local username.
            key_filename (str):
                Path to a private key for authentication. Default location will be used if omitted.
            password (str):
                Password for ssh authentication. Has lower precedence than private key.
            port (int):
                Optional port of the host.
            accept_unknown_host (bool):
                Whether to accept an unknown host key. Defaults to True.
        """
        self.hostname = hostname
        self.username = username
        self.key_filename = key_filename
        self.password = password
        if accept_unknown_host:
            self.missing_host_policy = paramiko.client.WarningPolicy()
        else:
            self.missing_host_policy = paramiko.client.RejectPolicy()
        self.port = port if port else 22

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """
        Run a command remotely via SSH.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (str):
                String to send to standard input for a command.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."
        with open_ssh_client() as client:
            client.load_system_host_keys()
            client.set_missing_host_key_policy(self.missing_host_policy)
            client.connect(
                self.hostname,
                username=self.username,
                port=self.port,
                password=self.password,
                key_filename=self.key_filename,
            )

            ssh_in, out, err = client.exec_command(cmd)
            if stdin:
                ssh_in.channel.send(stdin)
                ssh_in.channel.shutdown_write()

            out_text = out.read().decode("utf-8")
            err_text = err.read().decode("utf-8")
            if out.channel.recv_exit_status() != 0 and not tolerate_err:
                LOG.error("Command {0} failed with {1}".format(cmd, err_text))
                raise RuntimeError(err_msg)

        return out_text, err_text


class ContainerExecutor(Executor):
    """Run commands in a Docker container."""

    def __init__(
        self,
        image,
        base_url="unix://var/run/docker.sock",
        timeout=None,
        verify_tls=False,
        cert_path=None,
    ):
        """
        Initialize.

        Args:
            image (str):
                Path to an image which will be used for performing skopeo operations. Must be
                downloadable by Docker.
            base_url (str):
                Base URL of the Docker client.
            timeout (int):
                Default timeout for API calls, in seconds.
            verify_tls (bool):
                Whether to use TLS verification.
            cert_path (str|None):
                Custom path to TLS certificates. If not specified, '~/.docker' is used.
        """
        self.image = image

        kwargs = {}
        kwargs["base_url"] = base_url
        kwargs["version"] = "auto"
        if timeout:
            kwargs["timeout"] = timeout
        if verify_tls:
            kwargs["base_url"] = kwargs["base_url"].replace("tcp://", "https://")
            cert_path = cert_path or os.path.join(os.path.expanduser("~"), ".docker")
            if os.path.isdir(cert_path):
                kwargs["tls"] = docker.tls.TLSConfig(
                    client_cert=(
                        os.path.join(cert_path, "cert.pem"),
                        os.path.join(cert_path, "key.pem"),
                    ),
                    verify=os.path.join(cert_path, "ca.pem"),
                )

        self.client = APIClient(**kwargs)
        repo, tag = self.image.split(":", 1)
        self.client.pull(repo, tag=tag)
        self.container = self.client.create_container(self.image, detach=True, tty=True)
        self.client.start(self.container["Id"])

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Cleanup the container when used as a context manager."""
        self.client.remove_container(self.container["Id"], force=True)

    def _run_cmd(self, cmd, err_msg=None, tolerate_err=False, stdin=None):
        """
        Run a command locally.

        NOTE: Older versions of Docker API don't support demuxing of stdout and stderr.
        This means that data from both streams will be mixed together. To maintain compatibility
        with the other classes, same output will be returned twice as a tuple. Each string
        will contain the same mix of stdout and stderr messages.

        Args:
            cmd (str):
                Shell command to be executed.
            error_msg (str):
                Error message written when the command fails.
            tolerate_err (bool):
                Whether to tolerate a failed command.
            stdin (None|str):
                This parameter exists only for compatibility with parent class. Sending input
                to containers is not supported.

        Returns (str, str):
            Tuple of stdout and stderr generated by the command.
        """
        err_msg = err_msg or "An error has occured when executing a command."
        cmd_exec = self.client.exec_create(self.container["Id"], cmd)
        # Unfortunately, older versions of Docker API don't support demuxing stdout and stderr
        stdout = self.client.exec_start(cmd_exec["Id"])

        if self.client.exec_inspect(cmd_exec["Id"]).get("ExitCode") != 0 and not tolerate_err:
            LOG.error("Command {0} failed with {1}".format(cmd, stdout))
            raise RuntimeError(err_msg)

        if stdout is None:
            stdout = b""
        out_str = stdout.decode("utf-8")

        return (out_str, out_str)

    def _add_file(self, data, file_name):
        """
        Add a text file to the running container.

        The primary use-case is to store a secret which will be accessed from inside the container.
        File will be stored in the path /tmp/<file_name>.

        Args:
            data (str):
                Data that should be stored in the container.
            file_name (str):
                Name of the file.
        """
        data_stream = io.BytesIO()
        data_tar = tarfile.TarFile(fileobj=data_stream, mode="w")
        encoded_data = data.encode("utf-8")
        tarinfo = tarfile.TarInfo(name=file_name)
        tarinfo.size = len(encoded_data)
        tarinfo.mtime = time.time()
        data_tar.addfile(tarinfo, io.BytesIO(encoded_data))
        data_tar.close()

        data_stream.seek(0)
        success = self.client.put_archive(
            container=self.container["Id"], path="/tmp", data=data_stream
        )

        if not success:
            raise RuntimeError("File was not successfully added to the container")

    def skopeo_login(self, username=None, password=None):
        """
        Attempt to login to Quay if no login credentials are present.

        This method is reimplemented because it uses a different approach to input the password.

        Args:
            username (str):
                Username for login.
            password (str):
                Password for login.
        """
        cmd_check = "skopeo login --get-login quay.io"
        out, err = self._run_cmd(cmd_check, tolerate_err=True)
        if username and username in out:
            LOG.info("Already logged in to Quay.io")
            return

        if not username or not password:
            raise ValueError(
                "Skopeo login credentials are not present. Quay user and token must be provided."
            )
        LOG.info("Logging in to Quay with provided credentials")

        password_file = "skopeo_password.txt"
        self._add_file(password, password_file)

        cmd_login = (
            " sh -c 'cat /tmp/{1} | skopeo login --authfile $HOME/.docker/config.json"
            " -u {0} --password-stdin quay.io'"
        ).format(shlex_quote(username), shlex_quote(password_file))
        out, err = self._run_cmd(cmd_login)

        if "Login Succeeded" in out:
            LOG.info("Login successful")
        else:
            raise RuntimeError(
                "Login command didn't generate expected output. "
                "STDOUT: '{0}', STDERR: '{1}'".format(out, err)
            )



from collections import namedtuple
import logging

import requests

from .exceptions import BadPushItem, InvalidTargetSettings, InvalidRepository
from .utils.misc import run_entrypoint, get_internal_container_repo_name, log_step
from .quay_api_client import QuayApiClient
from .quay_client import QuayClient
from .container_image_pusher import ContainerImagePusher
from .signature_handler import ContainerSignatureHandler, OperatorSignatureHandler
from .signature_remover import SignatureRemover
from .operator_pusher import OperatorPusher
from .utils.misc import get_external_container_repo_name, get_pyxis_ssl_paths

# TODO: do we want this, or should I remove it?
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

LOG = logging.getLogger("pubtools.quay")


class PushDocker:
    """Handle full Docker push workflow."""

    ImageData = namedtuple("ImageData", ["repo", "tag", "digest"])

    def __init__(self, push_items, hub, task_id, target_name, target_settings):
        """
        Initialize.

        Args:
            push_items ([_PushItem]):
                List of push items.
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            task_id (str):
                task id
            target_name (str):
                Name of the target.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.hub = hub
        self.task_id = task_id
        self.target_name = target_name
        self.target_settings = target_settings

        self.verify_target_settings()

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")

        self._dest_quay_client = None
        self._dest_quay_api_client = None

    @property
    def dest_quay_client(self):
        """Create and access QuayClient for dest image."""
        if self._dest_quay_client is None:
            self._dest_quay_client = QuayClient(
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
                self.quay_host,
            )
        return self._dest_quay_client

    @property
    def dest_quay_api_client(self):
        """Create and access QuayApiClient for dest image."""
        if self._dest_quay_api_client is None:
            self._dest_quay_api_client = QuayApiClient(
                self.target_settings["dest_quay_api_token"], self.quay_host
            )
        return self._dest_quay_api_client

    def verify_target_settings(self):
        """Verify that target settings contains all the necessary data."""
        LOG.info("Verifying the necessary target settings")
        required_settings = [
            "source_quay_user",
            "source_quay_password",
            "dest_quay_user",
            "dest_quay_password",
            "dest_quay_api_token",
            "pyxis_server",
            "quay_namespace",
            "iib_krb_principal",
            "iib_organization",
            "iib_index_image",
            "quay_operator_repository",
            "skopeo_image",
            "docker_settings",
        ]
        for setting in required_settings:
            if setting not in self.target_settings:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the target settings.".format(setting)
                )

        required_docker_settings = ["umb_urls", "docker_reference_registry"]
        for setting in required_docker_settings:
            if setting not in self.target_settings["docker_settings"]:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the docker settings.".format(setting)
                )
        if (
            "iib_overwrite_from_index_token" in self.target_settings
            and "iib_overwrite_from_index" not in self.target_settings
        ) or (
            "iib_overwrite_from_index_token" not in self.target_settings
            and "iib_overwrite_from_index" in self.target_settings
        ):
            msg = (
                "Either both or neither of 'iib_overwrite_from_index' and "
                "'iib_overwrite_from_index_token' should be specified in target settings."
            )
            LOG.error(msg)
            raise InvalidTargetSettings(msg)

    @log_step("Get container push items")
    def get_docker_push_items(self):
        """
        Filter push items to only include docker ones.

        Also, Check the validity of these items and raise an exception in case of incorrect data.

        Returns ([ContainerPushItem]):
            Docker push items.
        """
        docker_push_items = []
        for item in self.push_items:
            if item.file_type != "docker":
                LOG.warning("Push item {0} doesn't have 'docker' type, skipping.".format(item))
                continue
            if item.errors:
                raise BadPushItem("Push item {0} contains errors: {1}".format(item, item.errors))
            if not item.metadata.get("pull_data"):
                raise BadPushItem("Push item {0} doesn't contain pull data.".format(item))
            LOG.info("Docker push item found: {0}".format(item))
            docker_push_items.append(item)

        return docker_push_items

    @log_step("Get operator push items")
    def get_operator_push_items(self):
        """
        Filter out push items to only include operator ones.

        Also, Check the validity of these items and raise an exception in case of incorrect data.

        Returns ([ContainerPushItem]):
            Operator push items.
        """
        operator_push_items = []
        for item in self.push_items:
            if item.file_type != "operator":
                LOG.warning("Push item {0} doesn't have 'operator' type, skipping.".format(item))
                continue
            if item.errors:
                raise BadPushItem("Push item {0} contains errors: {1}".format(item, item.errors))
            if not item.metadata.get("op_type"):
                raise BadPushItem("Push item {0} doesn't contain 'op_type'".format(item))
            if item.metadata["op_type"] == "appregistry":
                LOG.warning(
                    "Push item {0} is unsupported legacy (appregistry), skipping".format(item)
                )
                continue
            if item.metadata["op_type"] != "bundle":
                message = "Push item {0} has unknown op_type: '{1}'.".format(
                    item, item.metadata["op_type"]
                )
                raise BadPushItem(message)
            if not item.metadata.get("com.redhat.openshift.versions"):
                msg = "Push item {0} doesn't specify 'com.redhat.openshift.versions'".format(item)
                raise BadPushItem(msg)
            LOG.info("Operator push item found: {0}".format(item))
            operator_push_items.append(item)

        return operator_push_items

    @classmethod
    def get_repo_metadata(cls, repo, target_settings):
        """
        Invoke the 'get-repo-metadata' entrypoint from pubtools-pyxis.

        Args:
            repo (str):
                Repository to get the metadata of.
            target_settings (dict):
                Settings used for setting the values of the entrypoint parameters.

        Returns (dict):
            Parsed response from Pyxis.
        """
        cert, key = get_pyxis_ssl_paths(target_settings)

        args = ["--pyxis-server", target_settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--repo-name", repo]

        env_vars = {}
        metadata = run_entrypoint(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-repo-metadata"),
            "pubtools-pyxis-get-repo-metadata",
            args,
            env_vars,
        )
        return metadata

    @classmethod
    def check_repos_validity(cls, push_items, hub, target_settings):
        """
        Check if specified repos are valid and pushing to them is allowed.

        Specifically, this method checks if the repo exists in Comet and if it's not deprecated
        If pushing to prod, also check if the repo already exists in stage.

        Args:
            push_items ([ContainerPushItem]):
                Container push items containing the repositories.
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            target_settings (dict):
                Target settings.
        """
        repos = []
        for item in push_items:
            if item.external_repos:
                repos += item.external_repos.keys()
            else:
                repos += item.repos.keys()

        repos = sorted(list(set(repos)))
        repo_schema = "{namespace}/{repo}"

        # we'll need to get stage namespace from stage target settings
        if "propagated_from" in target_settings:
            stage_target_info = hub.worker.get_target_info(target_settings["propagated_from"])
            stage_namespace = stage_target_info["settings"]["quay_namespace"]
            stage_quay_client = QuayClient(
                stage_target_info["settings"]["dest_quay_user"],
                stage_target_info["settings"]["dest_quay_password"],
            )

        for repo in repos:
            LOG.info("Checking validity of Comet repository '{0}'".format(repo))
            # Check if repo exists in Comet
            try:
                metadata = cls.get_repo_metadata(repo, target_settings)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404:
                    raise InvalidRepository("Repository {0} doesn't exist in Comet".format(repo))
                else:
                    raise

            # Check if repo is not deprecated
            # TODO: check with Comet team if this is a reliable way of checking
            if "Deprecated" in metadata["release_categories"]:
                raise InvalidRepository("Repository {0} is deprecated".format(repo))

            # if we're pushing to prod target, check if repo exists on stage as well
            if "propagated_from" in target_settings:
                internal_repo = get_internal_container_repo_name(repo)
                full_repo = repo_schema.format(namespace=stage_namespace, repo=internal_repo)
                try:
                    stage_quay_client.get_repository_tags(full_repo)
                except requests.exceptions.HTTPError as e:
                    # strangely, if repo doesn't exist, 401 is returned if a robot account is used
                    if e.response.status_code == 404 or e.response.status_code == 401:
                        raise InvalidRepository(
                            "Repository {0} doesn't exist on stage".format(repo)
                        )
                    else:
                        raise

    @log_step("Generate backup mapping")
    def generate_backup_mapping(self, push_items):
        """
        Create resources which will be used for rollback if something goes wrong during the push.

        Specifically, create two resources: 'backup_tags' and 'rollback_tags'.
        - 'backup_tags' is a mapping of ImageData->manifest, and consists of images which will
          be overwritten. During rollback, tag is made to re-reference the old manifest.
        - 'rollback_tags' is a list of ImageData which don't yet exist. During rollback, they
          will be removed to preserve pre-push state.

        ImageData is a namedtuple used to assign and access parts of an image in a formatted way.

        Args:
            push_items ([ContainerPushItem]):
                Container push items.

        Returns (({ImageData: str}, [ImageData])):
            Tuple of backup_tags and rollback_tags
        """
        backup_tags = {}
        rollback_tags = []
        repo_schema = "{namespace}/{repo}"
        image_schema = "{host}/{repo}@{digest}"
        image_schema_tag = "{host}/{repo}:{tag}"
        namespace = self.target_settings["quay_namespace"]

        for item in push_items:
            for repo, tags in sorted(item.metadata["tags"].items()):
                internal_repo = get_internal_container_repo_name(repo)
                full_repo = repo_schema.format(namespace=namespace, repo=internal_repo)
                LOG.info("Generating backup mapping for repository '{0}'".format(repo))
                # try to get repo data
                try:
                    repo_tags = self.dest_quay_client.get_repository_tags(full_repo)
                except requests.exceptions.HTTPError as e:
                    # When robot account is used, 401 is returned instead of 404
                    if e.response.status_code == 404 or e.response.status_code == 401:
                        repo_tags = None
                    else:
                        raise

                for tag in tags:
                    # repo doesn't exist, add to rollback tags
                    if not repo_tags:
                        # for rollback tags digest is not known
                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None))
                        continue
                    # tag exists in the repo, add to backup tags
                    if tag in repo_tags.get("tags", {}):
                        image_tag = image_schema_tag.format(
                            host=self.quay_host, repo=full_repo, tag=tag
                        )
                        digest = self.dest_quay_client.get_manifest_digest(image_tag)
                        # for backup tags store also digest
                        image_data = PushDocker.ImageData(full_repo, tag, digest)
                        image = image_schema.format(
                            host=self.quay_host,
                            repo=full_repo,
                            digest=digest,
                        )
                        manifest = self.dest_quay_client.get_manifest(image)
                        backup_tags[image_data] = manifest
                    # tag doesn't exist in the repo, add to rollback tags
                    else:
                        # for rollback tags digest is not known
                        rollback_tags.append(PushDocker.ImageData(full_repo, tag, None))

        # it's possible that rollback tags will contain duplicate entries
        rollback_tags = sorted(list(set(rollback_tags)))
        return (backup_tags, rollback_tags)

    @log_step("Perform rollback")
    def rollback(self, backup_tags, rollback_tags):
        """
        Perform a rollback.

        Args:
            backup_tags ({ImageData: str}):
                Dictionary mapping of ImageData and a manifest before it was overwritten.
            rollback_tags ([ImageData]):
                List of newly added ImageData.
        """
        # restore overwritten tags to their original values
        schema = "{host}/{repo}:{tag}"
        LOG.info("Restoring tags to their original values")
        for image_data, manifest in sorted(backup_tags.items()):
            image_ref = schema.format(host=self.quay_host, repo=image_data.repo, tag=image_data.tag)
            LOG.info("Restoring tag '{0}'".format(image_ref))
            self.dest_quay_client.upload_manifest(manifest, image_ref)

        # delete tags that didn't previously exist
        LOG.info("Removing newly introduced tags")
        for image_data in rollback_tags:
            image_ref = schema.format(host=self.quay_host, repo=image_data.repo, tag=image_data.tag)
            LOG.info("Removing tag '{0}'".format(image_ref))
            try:
                self.dest_quay_api_client.delete_tag(image_data.repo, image_data.tag)
            except requests.exceptions.HTTPError as e:
                # if error occurred before tag was copied, it may not exist
                # deleting a non-existent tag shouldn't be an error
                if e.response.status_code != 404 and e.response.status_code != 401:
                    raise

    def remove_old_signatures(
        self,
        push_items,
        operator_push_items,
        existing_index_images,
        iib_results,
        backup_tags,
        container_signature_handler,
        operator_signature_handler,
        signature_remover,
    ):
        """
        Remove signatures of containers for tags which were overwritten in the current push.

        Method fetches all existing signatures for digests in backup tags. Only signatures with
        repo, tag pairs  matching to backup tags are then processes forward as also signatures
        for different repos could be returned. From those, only signature not
        matching signatures which were just created for current push are removed.
        Mechanism is the same for index images signatures, except those are not compared to
        any backup tags.

        Args:
            push_items ([_PushItem]):
                List of container push items.
            operator_push_items ([_PushItem]):
                List of operator push items.
            existing_index_images: ([(digest, version)]):
                List of tuple of digest + version(tag) of index images which existed
                before new index image was pushed in the current task
            iib_results (({str:dict})):
                Dictionary containing IIB results and signing keys for all OPM versions.
            backup_tags ({ImageData: str}):
                Dictionary of ImageData (repo, tag, digest) -> manifest
                holding containers which were overwritten in the currently running task
            container_signature_handler (ContainerSignatureHandler):
                ContanerSignatureHandler instance.
            operator_signature_handler (OperatorSignatureHandler):
                ContanerSignatureHandler instance.
            operator_signature_handler (SignatureRemover):
                SignatureRemover instance.
        """
        claim_messages = []
        for item in push_items:
            claim_messages += container_signature_handler.construct_item_claim_messages(item)
        new_signatures = [(m["manifest_digest"], m["docker_reference"]) for m in claim_messages]
        outdated_signatures = []

        for image_data, manifest in backup_tags.items():
            ext_repo = get_external_container_repo_name(image_data.repo.split("/")[1])
            if "manifests" in manifest:
                for arch_manifest in manifest["manifests"]:
                    outdated_signatures.append((arch_manifest["digest"], image_data.tag, ext_repo))
            else:
                outdated_signatures.append((image_data.digest, image_data.tag, ext_repo))

        signatures_to_remove = []
        for esig in container_signature_handler.get_signatures_from_pyxis(
            [sig[0] for sig in outdated_signatures]
        ):
            if (
                esig["manifest_digest"],
                esig["reference"].split(":")[-1],
                esig["repository"],
            ) in outdated_signatures and (
                esig["manifest_digest"],
                esig["reference"],
            ) not in new_signatures:
                signatures_to_remove.append(esig["_id"])

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        if signatures_to_remove:
            signature_remover.remove_signatures_from_pyxis(
                signatures_to_remove,
                self.target_settings["pyxis_server"],
                cert,
                key,
            )

        signatures_to_remove = []
        ii_claim_messages = []
        if existing_index_images:
            for version, iib_details in sorted(iib_results.items()):
                iib_result = iib_details["iib_result"]
                signing_keys = iib_details["signing_keys"]
                image_schema = "{host}/{namespace}/{repo}@{digest}"
                iib_namespace = iib_result.index_image_resolved.split("/")[1]
                image_digest = iib_result.index_image_resolved.split("@")[1]
                intermediate_index_image = image_schema.format(
                    host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
                    namespace=iib_namespace,
                    repo="iib",
                    digest=image_digest,
                )
                ii_claim_messages += (
                    operator_signature_handler.construct_index_image_claim_messages(
                        intermediate_index_image, version, signing_keys
                    )
                )
            new_operator_signatures = [
                (m["manifest_digest"], m["docker_reference"]) for m in ii_claim_messages
            ]

            for esig in container_signature_handler.get_signatures_from_pyxis(
                [digest_version[0] for digest_version in existing_index_images]
            ):
                if (esig["manifest_digest"], esig["reference"]) not in new_operator_signatures and (
                    esig["manifest_digest"],
                    esig["reference"].split(":")[-1],
                    esig["repository"],
                ) in existing_index_images:
                    signatures_to_remove.append(esig["_id"])
            if signatures_to_remove:
                signature_remover.remove_signatures_from_pyxis(
                    signatures_to_remove,
                    self.target_settings["pyxis_server"],
                    cert,
                    key,
                )

    def run(self):
        """
        Perform the full push-docker workflow.

        The workflow can be summarized as:
        - Filter out push items to only include container image items
        - Check if the destination repos may be pushed to (using Pyxis)
        - Generate backup mapping that will be used for rollback if something goes wrong.
        - Sign container images using RADAS and upload signatures to Pyxis
        - Push container images to their destinations
        - Filter out push items to only include operator image items
        - Add operator bundles to index images by using IIB
        - Sign index images using RADAS and upload signatures to Pyxis
        - Push the index images to Quay
        - Remove outdated container signatures
        - (in case of failure) Rollback destination repos to the pre-push state
        """
        # TODO: Do we need to manage push item state?
        # Filter out non-docker push items
        docker_push_items = self.get_docker_push_items()
        # Get operator push items (done early so that possible issues are detected)
        operator_push_items = self.get_operator_push_items()
        # Check if we may push to destination repos
        self.check_repos_validity(docker_push_items, self.hub, self.target_settings)
        # Generate resources for rollback in case there are errors during the push
        backup_tags, rollback_tags = self.generate_backup_mapping(docker_push_items)
        existing_index_images = []
        iib_results = None

        try:
            # Sign container images
            container_signature_handler = ContainerSignatureHandler(
                self.hub, self.task_id, self.target_settings, self.target_name
            )
            operator_signature_handler = OperatorSignatureHandler(
                self.hub, self.task_id, self.target_settings, self.target_name
            )
            sig_remover = SignatureRemover()
            sig_remover.set_quay_client(self.dest_quay_client)

            container_signature_handler.sign_container_images(docker_push_items)
            # Push container images
            container_pusher = ContainerImagePusher(docker_push_items, self.target_settings)
            container_pusher.push_container_images()

            if operator_push_items:
                # Build index images
                operator_pusher = OperatorPusher(operator_push_items, self.target_settings)
                existing_index_images = operator_pusher.get_existing_index_images(
                    self.dest_quay_client
                )
                iib_results = operator_pusher.build_index_images()
                # Sign operator images
                operator_signature_handler.sign_operator_images(iib_results)
                # Push index images to Quay
                operator_pusher.push_index_images(iib_results)
        except (Exception, SystemExit):
            LOG.error("An exception has occurred during the push, starting rollback")
            self.rollback(backup_tags, rollback_tags)
            raise
        else:
            # Remove old signatures
            self.remove_old_signatures(
                docker_push_items,
                operator_push_items,
                existing_index_images,
                iib_results,
                backup_tags,
                container_signature_handler,
                operator_signature_handler,
                sig_remover,
            )


def mod_entry_point(push_items, hub, task_id, target_name, target_settings):
    """Entry point for use in another python code."""
    push = PushDocker(push_items, hub, task_id, target_name, target_settings)
    push.run()



from collections import namedtuple
from copy import deepcopy
import json
import logging

import requests

from .command_executor import ContainerExecutor
from .exceptions import (
    BadPushItem,
    InvalidTargetSettings,
)
from .utils.misc import get_internal_container_repo_name, get_pyxis_ssl_paths
from .quay_client import QuayClient
from .container_image_pusher import ContainerImagePusher
from .signature_handler import SignatureHandler, BasicSignatureHandler
from .manifest_list_merger import ManifestListMerger
from .untag_images import untag_images
from .push_docker import PushDocker
from .signature_remover import SignatureRemover

# TODO: do we want this, or should I remove it?
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

LOG = logging.getLogger("pubtools.quay")


class TagDocker:
    """Handle full tag-docker workflow."""

    ImageDetails = namedtuple("ImageDetails", ["reference", "manifest", "manifest_type", "digest"])
    MANIFEST_LIST_TYPE = "application/vnd.docker.distribution.manifest.list.v2+json"
    MANIFEST_V2S2_TYPE = "application/vnd.docker.distribution.manifest.v2+json"

    def __init__(self, push_items, hub, task_id, target_name, target_settings):
        """
        Initialize.

        Args:
            push_items ([_PushItem]):
                List of push items.
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            task_id (str):
                task id
            target_name (str):
                Name of the target.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.hub = hub
        self.task_id = task_id
        self.target_name = target_name
        self.target_settings = target_settings

        self._quay_client = None

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")

        self.signature_remover = SignatureRemover()
        self.signature_remover.set_quay_client(self.quay_client)

        self.verify_target_settings()
        self.verify_input_data()

    @property
    def quay_client(self):
        """Create and access QuayClient for source and dest images."""
        if self._quay_client is None:
            self._quay_client = QuayClient(
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
                self.quay_host,
            )
        return self._quay_client

    def verify_target_settings(self):
        """Verify that target settings contains all the necessary data."""
        LOG.info("Verifying the necessary target settings")
        required_settings = [
            "source_quay_user",
            "source_quay_password",
            "dest_quay_user",
            "dest_quay_password",
            "dest_quay_api_token",
            "pyxis_server",
            "quay_namespace",
            "iib_index_image",
            "iib_krb_principal",
            "quay_operator_repository",
            "skopeo_image",
            "docker_settings",
        ]
        for setting in required_settings:
            if setting not in self.target_settings:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the target settings.".format(setting)
                )

        required_docker_settings = ["umb_urls", "docker_reference_registry"]
        for setting in required_docker_settings:
            if setting not in self.target_settings["docker_settings"]:
                raise InvalidTargetSettings(
                    "'{0}' must be present in the docker settings.".format(setting)
                )

    def verify_input_data(self):
        """Verify that the data specified for the TagDocker operation are correct."""
        LOG.info("Verifying the input data")
        for item in self.push_items:
            if item.file_type != "docker":
                raise BadPushItem("Push items must be of 'docker' type")
            if len(item.repos) != 1:
                raise BadPushItem("In tag-docker, push items must have precisely one repository.")
            if item.metadata["add_tags"] and not item.metadata["tag_source"]:
                raise BadPushItem("Source must be provided if tags were requested to be added.")
            if not item.metadata["new_method"]:
                raise BadPushItem("Only new method is supported for tag-docker in Quay.")
            if item.metadata["tag_source"] and ":" in item.metadata["tag_source"]:
                raise BadPushItem("Specifying source via digest is not allowed.")

    def check_input_validity(self):
        """
        Check if input data satisfies tag-docker specific constraints.

        The constraints are following:
        1. If adding tags to prod target, these tags must already exist in stage target.
        2. If removing tags from prod target, these tags must already not exist in stage target.
        """
        if "propagated_from" in self.target_settings:
            full_repo_schema = "{host}/{namespace}/{repo}"
            stage_target_info = self.hub.worker.get_target_info(
                self.target_settings["propagated_from"]
            )
            stage_namespace = stage_target_info["settings"]["quay_namespace"]
            stage_quay_client = QuayClient(
                stage_target_info["settings"]["dest_quay_user"],
                stage_target_info["settings"]["dest_quay_password"],
                self.quay_host,
            )

            for item in self.push_items:
                internal_repo = get_internal_container_repo_name(list(item.repos.keys())[0])
                stage_repo = full_repo_schema.format(
                    host=self.quay_host, namespace=stage_namespace, repo=internal_repo
                )

                # all to-be-added tags must already exist in stage repo
                for tag in item.metadata["add_tags"]:
                    stage_image = "{0}:{1}".format(stage_repo, tag)
                    try:
                        stage_quay_client.get_manifest(stage_image)
                    except requests.exceptions.HTTPError as e:
                        if e.response.status_code == 404 or e.response.status_code == 401:
                            raise BadPushItem(
                                "To-be-added tag {0} must already exist in stage repo".format(tag)
                            )
                        else:
                            raise

                # all to-be-removed tags must already be removed from stage
                for tag in item.metadata["remove_tags"]:
                    stage_image = "{0}:{1}".format(stage_repo, tag)
                    try:
                        stage_quay_client.get_manifest(stage_image)
                    except requests.exceptions.HTTPError as e:
                        if e.response.status_code == 404 or e.response.status_code == 401:
                            # 404/401 -> all good
                            pass
                        else:
                            raise
                    else:
                        raise BadPushItem(
                            "To-be-removed tag {0} must already be removed from stage repo".format(
                                tag
                            )
                        )

    def get_image_details(self, reference, executor):
        """
        Create an ImageDetails namedtuple for the given image reference.

        Args:
            reference (str):
                Image reference.
            executor (Executor):
                Instance of Executor subclass used for skopeo inspect.
        Returns (ImageDetails|None):
            Namedtuple filled with images data, or None if image doesn't exist.
        """
        LOG.info("Getting image details of {0}".format(reference))
        try:
            manifest = self.quay_client.get_manifest(reference)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404 or e.response.status_code == 401:
                LOG.info("Image '{0}' doesn't exist".format(reference))
                return None
            else:
                raise

        manifest_type = manifest["mediaType"]
        if manifest_type not in [TagDocker.MANIFEST_V2S2_TYPE, TagDocker.MANIFEST_LIST_TYPE]:
            raise BadPushItem("Image {0} has manifest type different than V2S2 or manifest list")

        # Check arch if the image is V2S2 manifest
        if manifest["mediaType"] == TagDocker.MANIFEST_V2S2_TYPE:
            arch = executor.skopeo_inspect(reference)["Architecture"]
            # Arch check is not a great way to verify that this is a source image, but there are
            # no better options without having build details
            if arch != "amd64":
                raise BadPushItem(
                    "Image {0} has V2S2 manifest and contains an architecture {1}. Only source "
                    "images are supported, which have arch 'amd64'.".format(reference, arch)
                )

        digest = self.quay_client.get_manifest_digest(reference)

        return TagDocker.ImageDetails(reference, manifest, manifest["mediaType"], digest)

    def is_arch_relevant(self, push_item, arch):
        """
        Find out if an operation should be performed on a given architecture.

        Uses values of 'archs' and 'exclude_archs' in push item's metadata.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            arch (str):
                Arch to investigate.
        Returns (bool):
            True if an operation should be performed on a given arch, False otherwise.
        """
        if push_item.metadata["exclude_archs"]:
            return arch not in push_item.metadata["archs"]
        else:
            return arch in push_item.metadata["archs"]

    def tag_remove_calculate_archs(self, push_item, tag, executor):
        """
        Calculate which architectures would be removed, and which would remain from a given tag.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag, for which a 'remove' operation will be performed.
            executor (Executor):
                Instance of Executor subclass used for skopeo inspect.
        Returns ([str], [str]):
            Tuple where first element contains archs that will be removed, and second element
            contains archs that will remain.
        """
        full_repo_schema = "{host}/{namespace}/{repo}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )

        if push_item.metadata["tag_source"]:
            source_image = "{0}:{1}".format(full_repo, push_item.metadata["tag_source"])
            source_details = self.get_image_details(source_image, executor)
        else:
            source_details = None

        dest_image = "{0}:{1}".format(full_repo, tag)
        dest_details = self.get_image_details(dest_image, executor)

        if dest_details is None:
            LOG.warning("Tag '{0}' already doesn't exist, no removal necessary".format(tag))
            return ([], [])

        if source_details and source_details.manifest_type != dest_details.manifest_type:
            raise BadPushItem(
                "Mismatch between manifest types of source {0}:{1} and tag {2}:{3}".format(
                    push_item.metadata["tag_source"],
                    source_details.manifest_type,
                    tag,
                    dest_details.manifest_type,
                )
            )

        # Scenario 1: source image
        if dest_details.manifest_type == TagDocker.MANIFEST_V2S2_TYPE:
            return self.tag_remove_calculate_archs_source_image(
                push_item, source_details, dest_details
            )

        # Scenario 2: multiarch image
        if dest_details.manifest_type == TagDocker.MANIFEST_LIST_TYPE:
            return self.tag_remove_calculate_archs_multiarch_image(
                push_item, source_details, dest_details
            )

    def tag_remove_calculate_archs_source_image(self, push_item, source_details, dest_details):
        """
        Calculate which archs would be removed if the specified images were source images.

        This method is a sub-step of the 'tag_remove_calculate_archs' method.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            source_details (ImageDetails|None):
                ImageDetails of source image, or None if it wasn't specified.
            dest_details (ImageDetails):
                ImageDetails of destination image.
        Returns ([str], [str]):
            Tuple where first element contains archs that will be removed, and second element
            contains archs that will remain.
        """
        # Option A: arch is relevant, source is specified and digests correspond -> remove
        if (
            self.is_arch_relevant(push_item, "amd64")
            and source_details is not None
            and source_details.digest == dest_details.digest
        ):
            return (["amd64"], [])
        # Option B: arch is relevant, source is specified, but digests don't correspond -> keep
        elif (
            self.is_arch_relevant(push_item, "amd64")
            and source_details is not None
            and source_details.digest != dest_details.digest
        ):
            return ([], ["amd64"])
        # Option C: arch is relevant, source is not specified (no digest check) -> remove
        elif self.is_arch_relevant(push_item, "amd64") and source_details is None:
            return (["amd64"], [])
        # Option D: arch is not relevant -> keep
        else:
            return ([], ["amd64"])

    def tag_remove_calculate_archs_multiarch_image(self, push_item, source_details, dest_details):
        """
        Calculate which archs would be removed if the specified images were multiarch images.

        This method is a sub-step of the 'tag_remove_calculate_archs' method.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            source_details (ImageDetails|None):
                ImageDetails of source image, or None if it wasn't specified.
            dest_details (ImageDetails):
                ImageDetails of destination image.
        Returns ([str], [str]):
            Tuple where first element contains archs that will be removed, and second element
            contains archs that will remain.
        """
        remove_archs = []
        keep_archs = []
        dest_manifest_data = [
            (m["digest"], m["platform"]["architecture"]) for m in dest_details.manifest["manifests"]
        ]
        source_manifest_data = (
            [
                (m["digest"], m["platform"]["architecture"])
                for m in source_details.manifest["manifests"]
            ]
            if source_details
            else []
        )

        for dest_digest, dest_arch in dest_manifest_data:
            # Option A: arch is relevant, src exists, digest matches in src and dest -> remove
            if (
                self.is_arch_relevant(push_item, dest_arch)
                and source_details is not None
                and (dest_digest, dest_arch) in source_manifest_data
            ):
                remove_archs.append(dest_arch)
            # Option B: arch is relevant, src exists, no digest match in src and dest -> keep
            elif (
                self.is_arch_relevant(push_item, dest_arch)
                and source_details is not None
                and (dest_digest, dest_arch) not in source_manifest_data
            ):
                keep_archs.append(dest_arch)
            # Option C: arch is relevant, src doesn't exist (digest match not possible)-> remove
            elif self.is_arch_relevant(push_item, dest_arch) and source_details is None:
                remove_archs.append(dest_arch)
            # Option D: arch is not relevant -> keep
            else:
                keep_archs.append(dest_arch)

        return (remove_archs, keep_archs)

    def tag_add_calculate_archs(self, push_item, tag, executor):
        """
        Calculate which architectures are present in a given tag, and which ones would be added.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag, for which an 'add' operation will be performed.
            executor (Executor):
                Instance of Executor subclass used for skopeo inspect.
        Returns ([str]|None):
            In case of multiarch image, arches which would be copied to the destination. In case
            of a source image, None if the copy operation is relevant or [] otherwise.
        """
        full_repo_schema = "{host}/{namespace}/{repo}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )
        source_image = "{0}:{1}".format(full_repo, push_item.metadata["tag_source"])
        dest_image = "{0}:{1}".format(full_repo, tag)
        source_details = self.get_image_details(source_image, executor)
        dest_details = self.get_image_details(dest_image, executor)

        if source_details is None:
            raise BadPushItem("Source image must be specified if add operation was requested")

        if dest_details and source_details.manifest_type != dest_details.manifest_type:
            raise BadPushItem(
                "Mismatch between manifest types of source {0}:{1} and tag {2}:{3}".format(
                    push_item.metadata["tag_source"],
                    source_details.manifest_type,
                    tag,
                    dest_details.manifest_type,
                )
            )

        # Scenario 1: source image
        if source_details.manifest_type == TagDocker.MANIFEST_V2S2_TYPE:
            # source arch is relevant, proceed with copying the source image
            if self.is_arch_relevant(push_item, "amd64"):
                return None
            # arch is irrelevant we want no-op
            else:
                return []

        # Scenario 2: multiarch image
        if source_details.manifest_type == TagDocker.MANIFEST_LIST_TYPE:
            add_archs = [
                m["platform"]["architecture"]
                for m in source_details.manifest["manifests"]
                if self.is_arch_relevant(push_item, m["platform"]["architecture"])
            ]
            return add_archs

    def copy_tag_sign_images(self, push_item, tag, signature_handler, executor):
        """
        Copy image from source to the destination tag and sign new manifest claims.

        If destination tag already contains a manifest, it will be overwritten.
        This workflow is expected to use on single-arch source images.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag, which acts as a destination to the copy operation.
            signature_handler (BasicSignatureHandler):
                Instance of signature handler which will perform the signing.
            executor (Executor):
                Instance of Executor subclass used for skopeo inspect.
        """
        full_repo_schema = "{host}/{namespace}/{repo}"
        external_image_schema = "{host}/{repo}:{tag}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )
        source_image = "{0}:{1}".format(full_repo, push_item.metadata["tag_source"])
        dest_image = "{0}:{1}".format(full_repo, tag)

        LOG.info(
            "Source image tag '{0}' will be copied to destination '{1}'".format(
                push_item.metadata["tag_source"], tag
            )
        )

        claim_messages = []
        details = self.get_image_details(source_image, executor)
        registries = self.target_settings["docker_settings"]["docker_reference_registry"]

        if details.manifest_type == TagDocker.MANIFEST_V2S2_TYPE and push_item.claims_signing_key:
            for registry in registries:
                reference = external_image_schema.format(
                    host=registry, repo=list(push_item.repos.keys())[0], tag=tag
                )
                message = SignatureHandler.create_manifest_claim_message(
                    destination_repo=list(push_item.repos.keys())[0],
                    signature_key=push_item.claims_signing_key,
                    manifest_digest=details.digest,
                    docker_reference=reference,
                    image_name=list(push_item.repos.keys())[0],
                    task_id=self.task_id,
                )
                claim_messages.append(message)
        elif details.manifest_type == TagDocker.MANIFEST_LIST_TYPE:
            raise ValueError("Tagging workflow is not supported for multiarch images")

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        self.signature_remover.remove_tag_signatures(
            reference=dest_image,
            pyxis_server=self.target_settings["pyxis_server"],
            pyxis_ssl_crtfile=cert,
            pyxis_ssl_keyfile=key,
            exclude_by_claims=claim_messages,
        )

        signature_handler.sign_claim_messages(claim_messages, True, True)
        ContainerImagePusher.run_tag_images(source_image, [dest_image], True, self.target_settings)

    def merge_manifest_lists_sign_images(self, push_item, tag, add_archs, signature_handler):
        """
        Merge manifest lists between source and destination tag and sign manifest claims.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag, which acts as a destination to the merge operation.
            add_archs ([str]):
                Architectures which should be copied to the existing manifest list.
            signature_handler (BasicSignatureHandler):
                Instance of signature handler which will perform the signing.
        """
        LOG.info(
            "Architectures {0} of tag '{1}' will be copied to destination tag '{2}'".format(
                add_archs, push_item.metadata["tag_source"], tag
            )
        )

        full_repo_schema = "{host}/{namespace}/{repo}"
        external_image_schema = "{host}/{repo}:{tag}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )
        source_image = "{0}:{1}".format(full_repo, push_item.metadata["tag_source"])
        dest_image = "{0}:{1}".format(full_repo, tag)

        claim_messages = []
        registries = self.target_settings["docker_settings"]["docker_reference_registry"]

        # NOTE: Arch images don't need to be copied, since they already exist in the same repo
        merger = ManifestListMerger(source_image, dest_image)
        merger.set_quay_clients(self.quay_client, self.quay_client)
        new_manifest_list = merger.merge_manifest_lists_selected_architectures(add_archs)

        if push_item.claims_signing_key:
            for manifest in new_manifest_list["manifests"]:
                for registry in registries:
                    reference = external_image_schema.format(
                        host=registry, repo=list(push_item.repos.keys())[0], tag=tag
                    )
                    message = SignatureHandler.create_manifest_claim_message(
                        destination_repo=list(push_item.repos.keys())[0],
                        signature_key=push_item.claims_signing_key,
                        manifest_digest=manifest["digest"],
                        docker_reference=reference,
                        image_name=list(push_item.repos.keys())[0],
                        task_id=self.task_id,
                    )
                    claim_messages.append(message)

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        # claim messages ensure that signatures of non-modified archs won't get removed
        self.signature_remover.remove_tag_signatures(
            reference=dest_image,
            pyxis_server=self.target_settings["pyxis_server"],
            pyxis_ssl_crtfile=cert,
            pyxis_ssl_keyfile=key,
            exclude_by_claims=claim_messages,
        )

        signature_handler.sign_claim_messages(claim_messages, True, True)

        raw_src_manifest = self.quay_client.get_manifest(source_image, manifest_list=True, raw=True)

        # Special case: if the source manifest and the merged manifest are the same, upload the
        # raw source manifest. The reason is that otherwise the digests of the copied manifests
        # will not correspond due to Python dicitonaries randomizing the element order
        if sorted(
            new_manifest_list["manifests"], key=lambda manifest: manifest["digest"]
        ) == sorted(
            json.loads(raw_src_manifest)["manifests"], key=lambda manifest: manifest["digest"]
        ):
            self.quay_client.upload_manifest(raw_src_manifest, dest_image, raw=True)
        else:
            self.quay_client.upload_manifest(new_manifest_list, dest_image)

    @classmethod
    def run_untag_images(cls, references, remove_last, target_settings):
        """
        Prepare the "untag images" entrypoint with all the necessary arguments and run it.

        Args:
            references ([str]):
                Image references which should be untagged.
            remove_last (bool):
                Whether to remove a tag when it's the last reference of an image (in that repo).
            target_settings (dict):
                Settings used for setting the value of untag parameters.
        """
        untag_images(
            references=references,
            quay_api_token=target_settings["dest_quay_api_token"],
            remove_last=remove_last,
            quay_user=target_settings["dest_quay_user"],
            quay_password=target_settings["dest_quay_password"],
            send_umb_msg=True,
            umb_urls=target_settings["docker_settings"]["umb_urls"],
            umb_cert=target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            # assumption that we'll continue using .pem format
            umb_client_key=target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            umb_ca_cert=target_settings["docker_settings"].get(
                "umb_ca_cert", "/etc/pki/tls/certs/ca-bundle.crt"
            ),
        )

    def untag_image(self, push_item, tag):
        """
        Untag image specified by tag.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag which should be removed.
        """
        LOG.info("Tag '{0}' will be removed".format(tag))
        full_repo_schema = "{host}/{namespace}/{repo}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )
        dest_image = "{0}:{1}".format(full_repo, tag)

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        self.signature_remover.remove_tag_signatures(
            reference=dest_image,
            pyxis_server=self.target_settings["pyxis_server"],
            pyxis_ssl_crtfile=cert,
            pyxis_ssl_keyfile=key,
        )

        self.run_untag_images([dest_image], True, self.target_settings)

    def manifest_list_remove_archs(self, push_item, tag, remove_archs):
        """
        Remove specified archs from a manifest list and upload a new manifest list to Quay.

        Args:
            push_item (ContainerPushItem):
                Push item to perform the workflow with.
            tag (str):
                Tag whose manifest's archs will be removed.
            remove_archs ([str]):
                Architectures to remove from the manifest list.
        """
        LOG.info("Architectures {0} of tag '{1}' will be removed".format(remove_archs, tag))
        full_repo_schema = "{host}/{namespace}/{repo}"
        namespace = self.target_settings["quay_namespace"]

        internal_repo = get_internal_container_repo_name(list(push_item.repos.keys())[0])
        full_repo = full_repo_schema.format(
            host=self.quay_host, namespace=namespace, repo=internal_repo
        )
        dest_image = "{0}:{1}".format(full_repo, tag)
        manifest_list = self.quay_client.get_manifest(dest_image, manifest_list=True)
        keep_manifests = []

        for manifest in manifest_list["manifests"]:
            if manifest["platform"]["architecture"] not in remove_archs:
                keep_manifests.append(deepcopy(manifest))

        new_manifest_list = deepcopy(manifest_list)
        new_manifest_list["manifests"] = keep_manifests

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        self.signature_remover.remove_tag_signatures(
            reference=dest_image,
            pyxis_server=self.target_settings["pyxis_server"],
            pyxis_ssl_crtfile=cert,
            pyxis_ssl_keyfile=key,
            remove_archs=remove_archs,
        )

        self.quay_client.upload_manifest(new_manifest_list, dest_image)

    def run(self):
        """
        Perform the full tag-docker workflow.

        The workflow adds or removes images to/from given tags. It's possible to specify only
        certain architectures to be added or removed. Based on the task arguments, five different
        scenarios may occur:
        - Copying whole image to tag (if tag is unused, or all its archs will be overwritten)
        - Copying only certain archs and adding them to the destination manifest list
        - Removing certain archs from a tag
        - Removing a tag (if all archs are to be removed from it)
        - No operation (to-be-removed tag already doesn't exist, or all archs are restricted)

        The workflow may be summarized as:
        - Verify that all repos may be worked with (same conditions are in PushDocker)
        - Evaluate which archs are to be added/removed from a given tag
        - If new images were created, perform signing workflow on them
        - Perform the appropriate add/remove/merge operation
        """
        # Validate repos, same as in PushDocker
        PushDocker.check_repos_validity(self.push_items, self.hub, self.target_settings)
        # perform tag-docker-specific checks
        self.check_input_validity()
        signature_handler = BasicSignatureHandler(self.hub, self.target_settings, self.target_name)

        with ContainerExecutor(
            self.target_settings["skopeo_image"],
            self.target_settings.get("docker_host") or "unix://var/run/docker.sock",
            self.target_settings.get("docker_timeout"),
            self.target_settings.get("docker_tls_verify") or False,
            self.target_settings.get("docker_cert_path") or None,
        ) as executor:
            executor.skopeo_login(
                self.target_settings["dest_quay_user"], self.target_settings["dest_quay_password"]
            )
            for item in self.push_items:
                for tag in item.metadata["add_tags"]:
                    LOG.info("Processing add tag '{0}'".format(tag))
                    add_archs = self.tag_add_calculate_archs(item, tag, executor)
                    # If all archs were somehow excluded from being added, no-op
                    if add_archs == []:
                        LOG.warning("No archs can be added to tag '{0}', skipping".format(tag))
                        continue
                    # If None, we're dealing with a source image and we want to copy to destination
                    elif add_archs is None:
                        self.copy_tag_sign_images(item, tag, signature_handler, executor)
                    # Otherwise, merge relevant archs of source and dest
                    else:
                        self.merge_manifest_lists_sign_images(
                            item, tag, add_archs, signature_handler
                        )

                for tag in item.metadata["remove_tags"]:
                    LOG.info("Processing remove tag '{0}'".format(tag))
                    remove_archs, keep_archs = self.tag_remove_calculate_archs(item, tag, executor)
                    # If all archs were somehow excluded from removal, no-op
                    if not remove_archs:
                        LOG.warning("No archs can be removed from tag '{0}', skipping".format(tag))
                        continue
                    # If no archs will remain after removal, just perform untagging
                    elif not keep_archs:
                        self.untag_image(item, tag)
                    # if some archs will be removed and some will remain, create new manifest list
                    else:
                        self.manifest_list_remove_archs(item, tag, remove_archs)


def mod_entry_point(push_items, hub, task_id, target_name, target_settings):
    """Entry point for use in another python code."""
    tag_docker = TagDocker(push_items, hub, task_id, target_name, target_settings)
    return tag_docker.run()



import logging

from .quay_client import QuayClient
from .quay_api_client import QuayApiClient

LOG = logging.getLogger("pubtools.quay")


class ImageUntagger:
    """Class containing logic for untagging images and deciding if they should be untagged."""

    def __init__(
        self,
        references,
        quay_api_token,
        remove_last=False,
        quay_user=None,
        quay_password=None,
        host=None,
    ):
        """
        Initialize.

        Args:
            references ([str]):
                List of image references to untag.
            quay_api_token (str):
                OAuth token for authentication of Quay REST API.
            remove_last (bool):
                Whether to remove a tag when it's the last reference of an image (in that repo).
            quay_user (str):
                Quay username for Docker HTTP API.
            quay_password (str):
                Quay password for Docker HTTP API.
            host (str):
                Custom Quay hostname (if required).
        """
        for reference in references:
            if "@" in reference:
                raise ValueError(
                    "Reference '{0}' must be specified via tag, not digest".format(reference)
                )

        if isinstance(host, str) and host[-1] == "/":
            host = host[:-1]
        self.host = host or "quay.io"

        self.references = references
        self.remove_last = remove_last

        if quay_user and quay_password:
            self._quay_client = QuayClient(quay_user, quay_password, self.host)
        else:
            self._quay_client = None
        self._quay_api_client = QuayApiClient(quay_api_token, self.host)

    def set_quay_client(self, quay_client):
        """
        Set client instance to be used for the HTTP API operations.

        Args:
            quay_client (QuayClient):
                Instance of QuayClient.
        """
        self._quay_client = quay_client

    def get_repository_tags_mapping(self):
        """
        Get a mapping of which tags would be removed from given repos based on provided refs.

        Returns ({str: [str]}):
            Mapping of repository->tags.
        """
        repo_tag_mapping = {}
        for reference in self.references:
            tag = reference.split(":")[-1]
            repo_path = reference.split(":")[0]
            # Take only last two parts of the url, which are namespace + repo
            repository = "/".join(repo_path.split("/")[-2:])
            repo_tag_mapping.setdefault(repository, []).append(tag)

        return repo_tag_mapping

    def construct_tag_digest_mappings(self, repository):
        """
        Create a mappings of tags->digests as well as digests->tags.

        Tags->digests mapping answers the question "Which digests does a given tag reference?".
        Digests->tags mapping answers the question "Which tags reference a given digest?".
        Tag may reference one digest (if image manifest), or multiple (if manifest list).

        Args:
            repository (str):
                Quay repository.

        Returns (({str: [str]}), ({str: [str]})):
            Tuple of dictionaries mapping tags->digests and digests->tags.
        """
        LOG.info("Gathering tags and digests of repository '{0}'".format(repository))
        tag_digest_mapping = {}
        digest_tag_mapping = {}
        image_schema = "{0}/{1}:{2}"
        repo_tags = self._quay_client.get_repository_tags(repository)

        for tag in repo_tags["tags"]:
            image = image_schema.format(self.host, repository, tag)
            manifest = self._quay_client.get_manifest(image)
            digest = self._quay_client.get_manifest_digest(image)

            # Option 1: No manifest list, only manifest
            if manifest["mediaType"] == QuayClient.MANIFEST_V2S2_TYPE:
                tag_digest_mapping[tag] = [digest]
                digest_tag_mapping.setdefault(digest, []).append(tag)

            # Option 2: We need to get digests of all architectures
            else:
                tag_digest_mapping[tag] = [digest]
                digest_tag_mapping.setdefault(digest, []).append(tag)

                for arch_manifest in manifest["manifests"]:
                    tag_digest_mapping[tag].append(arch_manifest["digest"])
                    digest_tag_mapping.setdefault(arch_manifest["digest"], []).append(tag)

        return (tag_digest_mapping, digest_tag_mapping)

    def get_lost_digests(self, tags, tag_digest_mapping, digest_tag_mapping):
        """
        Calculate a list of digests that would be lost if the provided tags were removed.

        Args:
            tags ([str]):
                List of tags that would be removed from a repo.
            tag_digest_mapping ({str: [str]}):
                Mapping of which digests are referenced by a given tag.
            digest_tag_mapping ({str: [str]}):
                Mapping of which tags reference a given digest.

        Returns ([str]):
            Digests that would be lost if given tags were removed.
        """
        remove_digests = []
        lost_digests = []
        for tag in tags:
            for digest in tag_digest_mapping.get(tag, []):
                if digest not in remove_digests:
                    remove_digests.append(digest)

        for digest in remove_digests:
            # which tags would remain referencing a given digest?
            remaining_tags = set(digest_tag_mapping[digest]) - set(tags)
            if len(remaining_tags) == 0:
                lost_digests.append(digest)

        return lost_digests

    def untag_images(self):
        """
        Determine if the specified tags may be removed and remove them.

        Returns ([str]):
            List of image references lost by the untagging.
        """
        if not self._quay_client:
            raise RuntimeError("QuayClient instance must be set")

        repo_tag_mapping = self.get_repository_tags_mapping()
        schema = "{0}/{1}@{2}"
        lost_imgs = []

        for repo in repo_tag_mapping:
            tag_digest_mapping, digest_tag_mapping = self.construct_tag_digest_mappings(repo)

            lost_digests = self.get_lost_digests(
                repo_tag_mapping[repo], tag_digest_mapping, digest_tag_mapping
            )
            if len(lost_digests) > 0:
                lost_imgs += [schema.format(self.host, repo, digest) for digest in lost_digests]

        if not self.remove_last and lost_imgs:
            raise ValueError(
                "Following images would no longer be referencable by tag after"
                " the untagging operation: '{0}'. Please specify --remove-last"
                " if this is not a concern.".format(lost_imgs)
            )

        if lost_imgs and self.remove_last:
            LOG.warning("Following images won't be referencable by tag: {0}".format(lost_imgs))
        if not lost_imgs:
            LOG.info("No images will be lost by this untagging operation")

        for reference in self.references:
            tag = reference.split(":")[-1]
            repo_path = reference.split(":")[0]
            # Take only last two parts of the url, which are namespace + repo
            repository = "/".join(repo_path.split("/")[-2:])
            LOG.info("Removing tag '{0}' from repository '{1}'".format(tag, repository))
            self._quay_api_client.delete_tag(repository, tag)

        return lost_imgs



# -*- coding: utf-8 -*-

"""setup.py"""

import os
import re
import sys

# import pkg_resources
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class Tox(TestCommand):
    user_options = [("tox-args=", "a", "Arguments to pass to tox")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import tox
        import shlex

        if self.tox_args:
            errno = tox.cmdline(args=shlex.split(self.tox_args))
        else:
            errno = tox.cmdline(self.test_args)
        sys.exit(errno)


def read_content(filepath):
    with open(filepath) as fobj:
        return fobj.read()


classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: Implementation :: PyPy",
]

# same dependency is called "docker-py" in for Python 2.6 and "docker" on newer Python versions
# Thus, it has to be removed from the list of requirements and added under a different name
PY26_BLACKLISTED_DEPENDENCIES = ["docker"]
PY26_EXTRA_DEPENDENCIES = ["docker-py"]

def building_rpm():
    """True when running within RPM build environment, which tweaks
    the build a little."""
    return 'RPM_PACKAGE_VERSION' in os.environ

def get_requirements():
    """
    Transform a list of requirements so that they are usable by older pip (9.0.0), and newer pip

    Regex extracts name and url from a tox-compatible format, and replaces it with only a name
    (which will be combined with dependency_links) or with PEP-508 compatible dependency.
    """
    with open("requirements.txt") as f:
        reqs = f.read().splitlines()
    
    if sys.version_info < (2, 7):
        reqs = [d for d in reqs if d not in PY26_BLACKLISTED_DEPENDENCIES]
        reqs.extend(PY26_EXTRA_DEPENDENCIES)

    # If we are building an RPM, we don't have pip available, and we want
    # to use the 'name + dependency_link' style
    if building_rpm():
        reqs = sorted(list(set(reqs)))
        pip_version = [0, 0, 0]
    else:
        import pip
        pip_version = [int(i) for i in pip.__version__.split('.')]

    reqs = [req for req in reqs if req != '']
    for i in range(len(reqs)):
        if pip_version < [19, 0, 0]:
            reqs[i] = re.sub(r'-e .*#egg=(.*)-.*', r'\1', reqs[i])
        else:
            reqs[i] = re.sub(r'-e (.*#egg=(.*)-.*)', r'\2 @ \1', reqs[i])

    return reqs


def get_dependency_links():
    """
    Extracts only depenency links for the dependency_links in older versions of pip.
    """
    with open("requirements.txt") as f:
        reqs = f.read().splitlines()
    dependency_links = []
    for req in reqs:
        link = re.subn(r'-e (.*#egg=.*-.*)', r'\1', req)
        if link[1] == 1:
            dependency_links.append(link[0])

    return dependency_links


long_description = read_content("README.rst") + read_content(
    os.path.join("docs/source", "CHANGELOG.rst")
)

extras_require = {"reST": ["Sphinx"]}
if os.environ.get("READTHEDOCS", None):
    extras_require["reST"].append("recommonmark")

setup(
    name="pubtools-quay",
    version="0.8.0",
    description="Pubtools-quay",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author="Lubomir Gallovic",
    author_email="lgallovi@redhat.com",
    url="https://github.com/release-engineering/pubtools-quay",
    classifiers=classifiers,
    packages=find_packages(exclude=["tests"]),
    data_files=[],
    install_requires=get_requirements(),
    dependency_links=get_dependency_links(),
    entry_points={
        "console_scripts": [
            "pubtools-quay-tag-image = pubtools._quay.tag_images:tag_images_main",
            "pubtools-quay-merge-manifest-list = "
            "pubtools._quay.merge_manifest_list:merge_manifest_list_main",
            "pubtools-quay-untag = pubtools._quay.untag_images:untag_images_main",
            "pubtools-quay-remove-repo = pubtools._quay.remove_repo:remove_repositories_main",
            "pubtools-quay-clear-repo = pubtools._quay.clear_repo:clear_repositories_main",
        ],
        "target": [
            "push-docker = pubtools._quay.push_docker:mod_entry_point",
            "tag-docker = pubtools._quay.tag_docker:mod_entry_point",
            "iib-add-bundles = pubtools._quay.iib_operations:iib_add_entrypoint",
            "iib-remove-operators = pubtools._quay.iib_operations:iib_remove_entrypoint",
            "iib-build-from-scratch = pubtools._quay.iib_operations:iib_from_scratch_entrypoint",
        ]
    },
    include_package_data=True,
    extras_require=extras_require,
    tests_require=["tox"],
    cmdclass={"test": Tox},
)



import logging

from .quay_session import QuaySession

LOG = logging.getLogger("pubtools.quay")


class QuayApiClient:
    """Class for performing Quay REST API queries."""

    def __init__(self, token, host=None):
        """
        Initialize.

        Args:
            token (str):
                Quay API token for authentication.
            host (str):
                Quay registry URL.
        """
        self.token = token
        self.session = QuaySession(hostname=host, api="quay")
        self.session.set_auth_token(self.token)

    def delete_tag(self, repository, tag):
        """
        Delete a tag from a repository.

        Args:
            repository (str):
                Repository in which the tag resides.
            tag (str):
                Tag to get its referenced images from.

        Returns (Response):
            Request library's Response object.
        """
        endpoint = "repository/{0}/tag/{1}".format(repository, tag)
        response = self.session.delete(endpoint)

        # Tag not existing is a tolerable error
        if "Invalid repository tag" not in response.text:
            response.raise_for_status()
        else:
            LOG.warning("Tag '{0}' already doesn't exist in repo '{1}'".format(tag, repository))

        return response

    def delete_repository(self, repository):
        """
        Delete a Quay repository.

        Args:
            repository (str):
                Repository to remove.

        Returns (Response):
            Request library's Response object.
        """
        endpoint = "repository/{0}".format(repository)
        response = self.session.delete(endpoint)

        # API doesn't return 404 even if repo doesn't exist. Thus no exception needs to be made
        response.raise_for_status()
        return response



import logging
import json
import tempfile

import requests

from .utils.misc import (
    get_internal_container_repo_name,
    run_entrypoint,
    get_external_container_repo_name,
)
from .quay_client import QuayClient

LOG = logging.getLogger("pubtools.quay")


class SignatureRemover:
    """Class used for finding the signatures that should be removed and removing them."""

    MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST = 50

    def __init__(self, quay_api_token=None, quay_user=None, quay_password=None, quay_host=None):
        """
        Initialize.

        Args:
            quay_user (str):
                User name for Quay Docker registry API.
            quay_password (str):
                Password for Quay Docker registry API.
            quay_host (str):
                Quay base host URL. Defaults to 'quay.io'.
        """
        self.quay_host = quay_host.rstrip("/") if quay_host else "quay.io"
        self.quay_user = quay_user
        self.quay_password = quay_password

        self._quay_client = None

    @property
    def quay_client(self):
        """Create and access QuayClient."""
        if self._quay_client is None:
            if not self.quay_user or not self.quay_password:
                raise ValueError(
                    "No instance of QuayClient is available. Please provide "
                    "'quay_user' and 'quay_password' or set the instance via 'set_quay_client'"
                )

            self._quay_client = QuayClient(self.quay_user, self.quay_password, self.quay_host)
        return self._quay_client

    def set_quay_client(self, quay_client):
        """
        Set a QuayClient instance.

        Args:
            quay_client (QuayClient):
                QuayClient instance.
        """
        self._quay_client = quay_client

    def get_signatures_from_pyxis(
        self,
        manifest_digests,
        pyxis_server,
        pyxis_ssl_crtfile,
        pyxis_ssl_keyfile,
    ):
        """
        Get existing signatures from Pyxis based on the specified criteria (currently only digests).

        NOTE: In the current implementation, only manifest digests are being used to search for
        existing signatures. Also, the search is performed in chunks, their size being limited by
        MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST.

        NOTE: This method is copied from SignatureHandler, although it doesn't utilize
        'target_settings' in order to be more versatile.

        Args:
            manifest_digests ([str]):
                Digests for which to return signatures.
            pyxis_server (str):
                URL of the Pyxis service.
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.

            Yields (dict):
                Existing signatures as returned by Pyxis based on specified criteria. The returned
                sturcture is an iterator to reduce memory requirements.
        """
        chunk_size = self.MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST
        manifest_digests = sorted(list(set(manifest_digests)))

        for chunk_start in range(0, len(manifest_digests), chunk_size):
            chunk = manifest_digests[chunk_start : chunk_start + chunk_size]  # noqa: E203

            args = ["--pyxis-server", pyxis_server]
            args += ["--pyxis-ssl-crtfile", pyxis_ssl_crtfile]
            args += ["--pyxis-ssl-keyfile", pyxis_ssl_keyfile]

            with tempfile.NamedTemporaryFile(
                mode="w", prefix="pubtools_quay_get_signatures_"
            ) as signature_fetch_file:
                if manifest_digests:
                    json.dump(chunk, signature_fetch_file)
                    signature_fetch_file.flush()
                    args += ["--manifest-digest", "@{0}".format(signature_fetch_file.name)]

                env_vars = {}
                chunk_results = run_entrypoint(
                    ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
                    "pubtools-pyxis-get-signatures",
                    args,
                    env_vars,
                )

            for result in chunk_results:
                yield result

    def remove_signatures_from_pyxis(
        self,
        signatures_to_remove,
        pyxis_server,
        pyxis_ssl_crtfile,
        pyxis_ssl_keyfile,
    ):
        """
        Remove signatures from Pyxis by using a pubtools-pyxis entrypoint.

        Args:
            signatures_to_remove ([str]):
                List of signature ids to be removed.
            pyxis_server (str):
                URL of the Pyxis service.
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.
        """
        LOG.info("Removing outdated signatures from pyxis")

        args = ["--pyxis-server", pyxis_server]
        args += ["--pyxis-ssl-crtfile", pyxis_ssl_crtfile]
        args += ["--pyxis-ssl-keyfile", pyxis_ssl_keyfile]

        with tempfile.NamedTemporaryFile(mode="w") as temp:
            json.dump(signatures_to_remove, temp)
            temp.flush()

            args += ["--ids", "@%s" % temp.name]

            env_vars = {}
            run_entrypoint(
                ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-delete-signatures"),
                "pubtools-pyxis-delete-signatures",
                args,
                env_vars,
            )

    def get_repository_digests(self, repository):
        """
        Get all digests of all images in a given repository.

        NOTE: Digests of manifest lists are not returned, as signing is not performed on them.

        Args:
            repository (str):
                Full Quay repository, including namespace.
        Returns ([str]):
            Digests of all images in a given repo.
        """
        image_schema = "{host}/{repo}:{tag}"
        digests = []
        repo_tags = self.quay_client.get_repository_tags(repository)

        for tag in repo_tags["tags"]:
            image = image_schema.format(host=self.quay_host, repo=repository, tag=tag)
            manifest = self.quay_client.get_manifest(image)
            # if V2S2, we want to include its digest
            if manifest["mediaType"] == "application/vnd.docker.distribution.manifest.v2+json":
                digests.append(self.quay_client.get_manifest_digest(image))
            else:
                for arch_manifest in manifest["manifests"]:
                    digests.append(arch_manifest["digest"])

        return sorted(list(set(digests)))

    def remove_repository_signatures(
        self,
        repository,
        namespace,
        pyxis_server,
        pyxis_ssl_crtfile,
        pyxis_ssl_keyfile,
    ):
        """
        Remove all signatures of all images in a given Quay repository.

        Args:
            repository (str):
                External name for a repository whose signatures should be removed.
            namespace (str):
                Quay namespace in which the repository resides.
            pyxis_server (str):
                URL of the Pyxis service.
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.
        """
        LOG.info("Removing signatures of all images of repository '{0}'".format(repository))

        internal_repo = "{0}/{1}".format(namespace, get_internal_container_repo_name(repository))
        remove_signature_ids = []
        digests = self.get_repository_digests(internal_repo)

        for signature in self.get_signatures_from_pyxis(
            digests,
            pyxis_server,
            pyxis_ssl_crtfile,
            pyxis_ssl_keyfile,
        ):
            if signature["repository"] == repository:
                remove_signature_ids.append(signature["_id"])

        if len(remove_signature_ids) > 0:
            LOG.info("{0} signatures will be removed".format(len(remove_signature_ids)))

            self.remove_signatures_from_pyxis(
                remove_signature_ids,
                pyxis_server,
                pyxis_ssl_crtfile,
                pyxis_ssl_keyfile,
            )
        else:
            LOG.info("No signatures need to be removed")

    def remove_tag_signatures(
        self,
        reference,
        pyxis_server,
        pyxis_ssl_crtfile,
        pyxis_ssl_keyfile,
        exclude_by_claims=None,
        remove_archs=None,
    ):
        """
        Remove signatures of an image specified by a tag.

        Source and multiarch images are supported. Signatures may be excluded from removal by
        specifying a list of claim messages. If existing signature and a claim messages matches,
        it is not removed, as the same exact signature would be recreated afterwards. Additionally,
        it's also possible to only specify certain architectures whose signatures will be removed.
        This option only has an effect when the image is a manifest list.

        Args:
            reference (str):
                Image reference whose signatures are to be removed.
            pyxis_server (str):
                URL of the Pyxis service.
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.
            exclude_by_claims ([dict]|None):
                List of claim messages whose existing signature matches will not be removed.
            remove_archs ([str]|None):
                If specified and the reference is a multiarch image, only signatures of given
                architectures will be eligible for removal.
        """
        if "@" in reference:
            raise ValueError("Image, whose signatures are being removed must be specified by tag.")

        full_repo, tag = reference.split(":", 1)
        external_repo = get_external_container_repo_name(full_repo.split("/")[-1])
        image_digests = []

        repo_tags = self.quay_client.get_repository_tags(full_repo.split("/", 1)[-1])
        # if specified tag doesn't exist in a repo, no-op
        if tag not in repo_tags["tags"]:
            return

        manifest = self.quay_client.get_manifest(reference)
        # V2S2 image, we need only manifest digest
        if manifest["mediaType"] == "application/vnd.docker.distribution.manifest.v2+json":
            image_digests.append(self.quay_client.get_manifest_digest(reference))
        # V2S2 image, we need digests of arch images
        else:
            for arch_manifest in manifest["manifests"]:
                if (
                    remove_archs is None
                    or arch_manifest["platform"]["architecture"] in remove_archs
                ):
                    image_digests.append(arch_manifest["digest"])

        new_claims_signatures = (
            list(set([(c["manifest_digest"], c["docker_reference"]) for c in exclude_by_claims]))
            if isinstance(exclude_by_claims, list)
            else []
        )

        remove_signature_ids = []
        for sig in self.get_signatures_from_pyxis(
            image_digests,
            pyxis_server,
            pyxis_ssl_crtfile,
            pyxis_ssl_keyfile,
        ):
            # if signature corresponds to to-be-removed digest+reference and isn't among new sigs
            if (
                sig["manifest_digest"] in image_digests
                and sig["repository"] == external_repo
                and sig["reference"].split(":")[-1] == tag
                and (sig["manifest_digest"], sig["reference"]) not in new_claims_signatures
            ):
                remove_signature_ids.append(sig["_id"])

        if len(remove_signature_ids) > 0:
            LOG.info("{0} signatures will be removed".format(len(remove_signature_ids)))

            self.remove_signatures_from_pyxis(
                remove_signature_ids,
                pyxis_server,
                pyxis_ssl_crtfile,
                pyxis_ssl_keyfile,
            )
        else:
            LOG.info("No signatures need to be removed")

    def get_index_image_signatures(
        self,
        image,
        claim_messages,
        pyxis_server,
        pyxis_ssl_crtfile,
        pyxis_ssl_keyfile,
    ):
        """
        Get existing signatures of an index image.

        NOTE: Image is expected to be in an internal format.

        Args:
            image (str):
                Image, whose signatures should be gathered.
            claim_messages (str):
                Newly constructed claim messages used for excluding matching signatures. Although,
                digests should never match for a new index image, it's added just in case.
            pyxis_server (str):
                URL of the Pyxis service.
            pyxis_ssl_crtfile (str):
                Path to .crt file for SSL authentication.
            pyxis_ssl_keyfile (str):
                Path to .key file for SSL authentication.
        Returns ([dict]):
            Existing signatures of the index image.
        """
        if "@" in image:
            raise ValueError("Please specify the index image via tag")

        # We'll assume manifest list, since it's an index image
        try:
            manifest = self.quay_client.get_manifest(image, manifest_list=True)
        except requests.exceptions.HTTPError as e:
            # Perhaps destination index image doesn't exist, tolerate 404
            if e.response.status_code == 404 or e.response.status_code == 401:
                return []
            else:
                raise

        digests = [m["digest"] for m in manifest["manifests"]]
        matched_signatures = []
        repo, tag = image.split(":", 1)
        external_repo = get_external_container_repo_name(repo.split("/")[-1])
        claims_by_key = [
            (c["manifest_digest"], c["repo"], c["docker_reference"].split(":", 1)[-1])
            for c in claim_messages
        ]

        for signature in self.get_signatures_from_pyxis(
            digests,
            pyxis_server,
            pyxis_ssl_crtfile,
            pyxis_ssl_keyfile,
        ):
            # if signature matches the old index image and isn't among new claims, it can be removed
            if (
                signature["manifest_digest"] in digests
                and signature["reference"].split(":", 1)[-1] == tag
                and signature["repository"] == external_repo
                and (
                    signature["manifest_digest"],
                    signature["repository"],
                    signature["reference"].split(":", 1)[-1],
                )
                not in claims_by_key
            ):
                matched_signatures.append(signature)

        return matched_signatures



import logging

from pubtools.pluggy import pm, task_context

from .signature_remover import SignatureRemover
from .quay_api_client import QuayApiClient
from .utils.misc import (
    setup_arg_parser,
    add_args_env_variables,
    send_umb_message,
    get_internal_container_repo_name,
)

LOG = logging.getLogger("pubtools.quay")

REMOVE_REPO_ARGS = {
    ("--repositories",): {
        "help": "External repositories to remove as CSV.",
        "required": True,
        "type": str,
    },
    ("--quay-org",): {
        "help": "Quay organization in which repositories reside.",
        "required": True,
        "type": str,
    },
    ("--quay-api-token",): {
        "help": "OAuth token for Quay REST API.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_API_TOKEN",
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": True,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--pyxis-server",): {
        "help": "Pyxis service hostname",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-crtfile",): {
        "help": "Path to .crt file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--pyxis-ssl-keyfile",): {
        "help": "Path to .key file for the SSL authentication",
        "required": True,
        "type": str,
    },
    ("--send-umb-msg",): {
        "help": "Flag of whether to send a UMB message",
        "required": False,
        "type": bool,
    },
    ("--umb-url",): {
        "help": "UMB URL. More than one can be specified.",
        "required": False,
        "type": str,
        "action": "append",
    },
    ("--umb-cert",): {
        "help": "Path to the UMB certificate for SSL authentication.",
        "required": False,
        "type": str,
    },
    ("--umb-client-key",): {
        "help": "Path to the UMB private key for accessing the certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-ca-cert",): {
        "help": "Path to the UMB CA certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-topic",): {
        "help": "UMB topic to send the message to.",
        "required": False,
        "type": str,
        "default": "VirtualTopic.eng.pub.quay_remove_repositories",
    },
}


def construct_kwargs(args):
    """
    Construct a kwargs dictionary based on the entered command line arguments.

    Args:
        args (argparse.Namespace):
            Parsed command line arguments.

    Returns (dict):
        Keyword arguments for the 'remove_repository' function.
    """
    kwargs = args.__dict__

    # in args.__dict__ unspecified bool values have 'None' instead of 'False'
    for name, attributes in REMOVE_REPO_ARGS.items():
        if attributes["type"] is bool:
            bool_var = name[0].lstrip("-").replace("-", "_")
            if kwargs[bool_var] is None:
                kwargs[bool_var] = False

    # some exceptions have to be remapped
    kwargs["umb_urls"] = kwargs.pop("umb_url")

    return kwargs


def verify_remove_repo_args(send_umb_msg, umb_urls, umb_cert):
    """
    Verify the presence and correctness of input parameters.

    Args:
        send_umb_msg (bool):
            Whether to send UMB messages about the untagged images.
        umb_urls ([str]):
            AMQP broker URLs to connect to.
        umb_cert (str):
            Path to a certificate used for UMB authentication.
    """
    if send_umb_msg:
        if not umb_urls:
            raise ValueError("UMB URL must be specified if sending a UMB message was requested.")
        if not umb_cert:
            raise ValueError(
                "A path to a client certificate must be provided when sending a UMB message."
            )


def remove_repositories(
    repositories,
    quay_org,
    quay_api_token,
    quay_user,
    quay_password,
    pyxis_server,
    pyxis_ssl_crtfile,
    pyxis_ssl_keyfile,
    send_umb_msg=False,
    umb_urls=[],
    umb_cert=None,
    umb_client_key=None,
    umb_ca_cert=None,
    umb_topic="VirtualTopic.eng.pub.quay_remove_repository",
):
    """
    Remove Quay repository.

    Args:
        repositories (str):
            External repositories to remove. Comma separated values.
        quay_org (str):
            Quay organization in which repositories reside.
        quay_api_token (str):
            OAuth token for authentication of Quay REST API.
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
        pyxis_server (str):
            Pyxis service hostname:
        pyxis_ssl_crtfile (str):
            Path to .crt file for SSL authentication.
        pyxis_ssl_keyfile (str):
            Path to .key file for SSL authentication.
        send_umb_msg (bool):
            Whether to send UMB messages about the untagged images.
        umb_urls ([str]):
            AMQP broker URLs to connect to.
        umb_cert (str):
            Path to a certificate used for UMB authentication.
        umb_client_key (str):
            Path to a client key to decrypt the certificate (if necessary).
        umb_ca_cert (str):
            Path to a CA certificate (for mutual authentication).
        umb_topic (str):
            Topic to send the UMB messages to.
    """
    parsed_repositories = repositories.split(",")
    verify_remove_repo_args(send_umb_msg, umb_urls, umb_cert)

    LOG.info("Removing repositories '{0}'".format(repositories))
    quay_api_client = QuayApiClient(quay_api_token)

    sig_remover = SignatureRemover(quay_user=quay_user, quay_password=quay_password)

    for repository in parsed_repositories:
        sig_remover.remove_repository_signatures(
            repository,
            quay_org,
            pyxis_server,
            pyxis_ssl_crtfile,
            pyxis_ssl_keyfile,
        )

        internal_repo = "{0}/{1}".format(quay_org, get_internal_container_repo_name(repository))
        quay_api_client.delete_repository(internal_repo)

    LOG.info("Repositories have been removed")
    pm.hook.quay_repositories_removed(repository_ids=sorted(parsed_repositories))

    if send_umb_msg:
        LOG.info("Sending a UMB message")
        props = {"removed_repositories": parsed_repositories}
        send_umb_message(
            umb_urls,
            props,
            umb_cert,
            umb_topic,
            client_key=umb_client_key,
            ca_cert=umb_ca_cert,
        )


def setup_args():
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(REMOVE_REPO_ARGS)


def remove_repositories_main(sysargs=None):
    """Entrypoint for removing repositories."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, REMOVE_REPO_ARGS)

    if not args.quay_api_token:
        raise ValueError("--quay-api-token must be specified")
    if not args.quay_password:
        raise ValueError("--quay-password must be specified")

    kwargs = construct_kwargs(args)

    with task_context():
        remove_repositories(**kwargs)
